import React = require('react');
import query = require("./query");
import promiseEx = require("./promiseEx");
import cargando = require("./cargando");

export interface AsyncProps<T> {
    /**Promesa que al ser resuelta devuelve el valor que se utilizará para dibujar el control */
    value: PromiseLike<T> | T;
    /**Dibuja el componente dado el resultado de la promesa */
    map: (value: T) => JSX.Element | null;
    /**Componente que se va a mostrar cuando se este cargando, si no se define se usara el spinner */
    cargando?: JSX.Element | null;

    /**Componente que se va a mostrar cuando se este en estado de error, si no se define se usara un label rojo */
    error?: (error: string | number) => JSX.Element | null;
    /**Tiempo que se debe de esperar antes de poner el cargando en milisegundos. Por default son 500 ms */
    delay?: number;
}

export interface AsyncState<T> {
    /**Resultado del query*/
    Model?: T | null;
    /**True si se estan cargando los elementos */
    Cargando?: boolean;
    /**Si hubo un error al cargar se mostrará aqui */
    Error?: string | number | null;
}

/**Devuelve true si un objeto es una promesa */
function IsPromise(value: any) {
    return value != null && (typeof value.then === 'function');
}

const disableObsoleteLog = "_disableObsoleteLog";

/**Muestra un componente que depende del valor de una promesa */
export class Async extends React.PureComponent<AsyncProps<any>, AsyncState<any>> {
    constructor(props, ) {
        super(props);
        this.state = { Cargando: true };

        if (!(props as any)[disableObsoleteLog]) {
            console.warn("Utilice la función async en lugar del componente Async, ya que la función esta tipada");
        }
    }


    componentWillMount() {

        this._isMounted = true;
        //Actualizamos el valor
        this.updateValue(this.props.value);
    }
    updateValue(nextValue: PromiseLike<any> | any) {
        if (IsPromise(nextValue)) {
            this.refreshPromise(nextValue);
        } else {
            this.setState({ Model: nextValue, Cargando: false, Error: null });
        }
    }

    componentWillReceiveProps(nextProps: AsyncProps<any>) {
        const nextValue = nextProps.value;
        const lastValue = this.props.value;
        if (nextValue != lastValue) {
            this.updateValue(nextProps.value);
        }
    }

    _isMounted: boolean;
    componentWillUnmount() {
        this._isMounted = false;
    }
    private currentVersion = 0;
    refreshPromise = async (promise: PromiseLike<any>) => {
        try {
            const currentVersion = this.currentVersion + 1;
            this.currentVersion = currentVersion;

            const minDelay = this.props.delay == null ? 500 : this.props.delay;
            //Si la promesa termina antes de minDelay ms no mostramos el cargando
            let race = await Promise.race([promiseEx.Delay(minDelay).then(() => "_delay"), promise]);
            if (race == "_delay") {
                //Si termino primero el delay, mostramos el cargando
                const stateChange = { Cargando: true, Error: null };
                this.setState(prev => this.currentVersion == currentVersion ? stateChange : {});
            }

            var result = await promise;
            if (this._isMounted) {
                const stateChange = {
                    Model: result,
                    Cargando: false,
                    Error: null
                };
                this.setState(prev => this.currentVersion == currentVersion ? stateChange : {})
            }
        }
        catch (err) {
            console.error("Async component error: ");
            console.error(err);

            if (this._isMounted) {
                this.setState({
                    Model: null,
                    Cargando: false,
                    Error: err
                });
            }
        }
    }

    get CargandoComponent(): JSX.Element | null {
        return (this.props.cargando !== undefined) ? this.props.cargando : <cargando.Spinner />
    }

    get ErrorComponent(): JSX.Element | null {
        return (this.props.error !== undefined) ? this.props.error(this.state.Error!) : <label style={{ color: "red" }}>{"Error: " + this.state.Error}</label>
    }

    render() {
        const childComponent = (!this.state.Error && !this.state.Cargando) &&
            this.props.map(this.state.Model);

        return this.state.Error != null ? this.ErrorComponent :
            this.state.Cargando ? this.CargandoComponent :
                (childComponent || null)
    }
}

/**Muestra un componente basado en el resultado de una promesa
 * @param value Promesa que se va a resolver para mostrar el componente
 * @param map Función que toma el resultado de la promesa y devuelve el componente que se desea dibujar
 */
export function async<T>(value: T | PromiseLike<T>, map: (value: T) => JSX.Element | null, options?: {
    /**Componente que se va a mostrar cuando se este cargando, si no se define se usara el spinner */
    cargando?: JSX.Element | null;

    /**Componente que se va a mostrar cuando se este en estado de error, si no se define se usara un label rojo */
    error?: (error: string | number) => JSX.Element | null;
    /**Tiempo que se debe de esperar antes de poner el cargando en milisegundos. Por default son 500 ms */
    delay?: number;
}) {
    const props = {
        ...(options || {}),
        value: value,
        map: map,
        [disableObsoleteLog]: true,
    } as AsyncProps<any>;

    return <Async {...props} />
}

export type PromiseValue<T> = { value: T, status: "done" } | { status: "pending" } | { status: "error", error: any };
/**
 * Muestra un componente basado en el resultado ya sea resuelto, pendiente o erroneo de una promesa
 * @param value Promesa
 * @param map Funcion que mapea el estado de la promesa a un componente de real
 */
export function asyncProps<T>(value: T | PromiseLike<T>, map: (value: PromiseValue<T>) => JSX.Element | null) {
    return async(value, x => map({ status: "done", value: x }), {
        cargando: map({ status: "pending" }),
        error: err => map({ status: "error", error: err })
    });
}

/**Crea una nueva promesa y devuelve por separado la promesa y las funciones que resuelven y rechazan a la promesa */
export function splitPromise<T>(): { promise: Promise<T>, resolve: (value?: T | PromiseLike<T> | undefined) => void, reject: (reason?: any) => void } {
    let resolve = null as any;
    let reject = null as any;
    const promise = new Promise<T>((promResolve, promReject) => {
        resolve = promResolve;
        reject = promReject;
    });

    return { promise, resolve, reject };
}

/**Crea una promesa que se resuelve en cierto tiempo */
export function delay<T>(milliseconds: number) {
    return new Promise<void>(resolve => window.setTimeout(resolve, milliseconds));
}

/**Aplica una función a cada propiedad de un objeto y devuelve un nuevo objeto con las mismas propiedades pero transformadas por la funcion dada */
function mapObject<T>(obj: T, map: (value: T[keyof T], key: keyof T) => any) {
    const ret: {[P in keyof T]: any } = {} as any;
    for (const key in obj) {
        ret[key] = map(obj[key], key);
    }
    return ret;
}
