
import React = require('react');
import async = require('./async');
import cargando = require("./cargando");
import { Delay } from "./promiseEx";
import swal = require("sweetalert");
import { componentToRx, Rxfy } from "react-rxk";
import { toObservable } from "keautils";
import * as rx from "rxjs";
/**Propiedades del botón */
export interface BaseButtonProps {
    /**Accion que se ejecuta cuando el usuario presiona el botón guardar */
    onClick?: () => PromiseLike<void> | void;
    disabled?: boolean | Promise<boolean> | rx.Observable<boolean>;
    tooltip?: string;
    block?: boolean;
}

export interface PromiseActionProps {
    onClick?: () => PromiseLike<void> | void;
    buttonFactory: (onClick?: () => void) => JSX.Element | null;
    cargandoComponent?: JSX.Element | null;
    errorFactory?: (error: string | number, onClick?: () => void) => JSX.Element | null;
}

export interface PromiseButtonState {
    Promise?: PromiseLike<void> | void;
}

/**Boton que acepta un metodo asíncrono en el on-click y que se deshabilita mientras se esta
 * ejecutando la promesa
 */
export class PromiseAction extends React.PureComponent<PromiseActionProps, PromiseButtonState> {
    constructor(props) {
        super(props);
        this.state = { Promise: undefined };
    }

    async click() {
        try {
            if (this.props.onClick)
                await this.props.onClick()
        } catch (error) {
            //Espera que evita un bug del swal, si dos swals se lanzan uno seguido despues del otro la UI se queda pegada
            await Delay(100);

            if (error == 401) {
                await new Promise(resolve => swal({
                    title: "No autorizado",
                    text: "Su usuario no esta autorizado para realizar esta operación",
                    type: "error"
                }, resolve));
            } else {
                await new Promise(resolve => swal({
                    title: "Error en la operación",
                    text: "Hubo un problema con la operación",
                    type: "error"
                }, resolve));
            }

            throw error;
        }
    }

    handleRequest = async () => {
        this.setState({ Promise: this.click() });
    }

    map = () => this.props.buttonFactory(this.handleRequest);

    render() {
        const error = this.props.errorFactory && ((error) => this.props.errorFactory!(error, this.handleRequest));
        return (
            async.async(
                this.state.Promise, this.map, {
                    cargando: this.props.cargandoComponent,
                    error,
                    delay: 0
                }
            )
        );
    }
}

export type BootstrapButtonStyle = "default" | "primary" | "success" | "info" | "warning" | "danger" | "link";

export interface ButtonProps extends BaseButtonProps {
    /**Estilo del botón de bootstrap. Por default es "default" */
    style?: BootstrapButtonStyle;
    /**True para poner el estilo de block en el botón, lo que hace que ocupe todo el espacio de la celda */
    block?: boolean;
    /**Clase del icono o indefinido para no mostrar un icono */
    iconClass?: string;
};

/**Vista del botón de bootstrap, no soporta promesa */
class BootstrapButtonSync extends React.PureComponent<ButtonProps, {}>
{
    render() {
        const className =
            "btn btn-" +
            (this.props.style || "default") +
            (this.props.disabled ? " disabled" : "") +
            (this.props.block ? " btn-block" : "");

        return (
            <button
                type="button"
                className={className}
                title={this.props.tooltip}
                onClick={this.props.disabled ? undefined : this.props.onClick!} >
                {this.props.iconClass && <i className={this.props.iconClass} />} {this.props.children}
            </button>
        );
    }
}


class ButtonNoRx extends React.PureComponent<ButtonProps, {}> {
    errorFactory = (error: any, onClick: () => void) =>
        <BootstrapButtonSync
            style="danger"
            block={this.props.block}
            onClick={onClick}
            iconClass="fa fa-times"
            children="Error"
            tooltip={error}
        />;

    buttonFactory = (onClick, cargando: boolean) => <BootstrapButtonSync
        children={this.props.children}
        disabled={this.props.disabled || cargando}
        iconClass={cargando ? "fa fa-circle-o-notch fa-spin" : this.props.iconClass}
        onClick={onClick}
        style={this.props.style}
        block={this.props.block}
        tooltip={this.props.tooltip} />;

    render() {
        return (
            <PromiseAction
                buttonFactory={onClick => this.buttonFactory(onClick, false)}
                errorFactory={this.errorFactory}
                cargandoComponent={this.buttonFactory(null, true)}
                onClick={this.props.onClick}
            />
        );
    }
}

const ButtonRx = componentToRx(ButtonNoRx);
export const Button = (props: Rxfy<ButtonProps>) => <ButtonRx {...props} disabled={toObservable(props.disabled).startWith(true)} />

export class BootstrapButton extends React.PureComponent<ButtonProps, {}> {
    constructor(props) {
        super(props);
        console.warn("BootstrapButton esta obsoleto. Utilice Button");
    }

    render() {
        return React.createElement(Button, this.props, this.props.children);
    }
}

/**Botón de agregar*/
export class AgregarButton extends React.PureComponent<BaseButtonProps, {}> {
    render() {
        return (
            <Button
                disabled={this.props.disabled}
                onClick={this.props.onClick}
                block={this.props.block}
                style="primary"
                iconClass="fa fa-plus" >
                Agregar
             </Button>
        );
    }
}

/**Boton de guardar*/
export class GuardarButton extends React.PureComponent<BaseButtonProps, {}> {
    render() {
        return (
            <Button
                disabled={this.props.disabled}
                onClick={this.props.onClick}
                block={this.props.block}
                style="success"
                iconClass="fa fa-floppy-o"  >
                Guardar
             </Button>
        );
    }
}

/**Boton de borrar*/
export class BorrarButton extends React.PureComponent<BaseButtonProps, {}> {
    render() {
        return (
            <Button
                disabled={this.props.disabled}
                onClick={this.props.onClick}
                block={this.props.block}
                style="danger"
                iconClass="fa fa-trash-o"  >
                Borrar
             </Button>
        );
    }
}


/**Boton de cancelar*/
export class CancelarButton extends React.PureComponent<BaseButtonProps, {}> {
    render() {
        return (
            <Button
                disabled={this.props.disabled}
                onClick={this.props.onClick}
                block={this.props.block}
                style="default"
                iconClass="fa fa-ban"  >
                Cancelar
             </Button>
        );
    }
}

/**Botón para buscar */
export class BuscarButton extends React.PureComponent<BaseButtonProps> {
    render() {
        return (
            <Button {... this.props} style="info" iconClass="fa fa-search" >
                Buscar
            </Button>
        )
    }
}

/**Boton para ir hacia atras en la navegación, si no se establece el on click, viaja hacia atras en la navegación
 * por default
 */
export class RegresarButton extends React.PureComponent<{ onClick?: () => void | Promise<void>, block?: boolean }, {}> {
    handleOnClick = () => {
        window.history.back();
    }
    render() {
        return (
            <Button
                onClick={this.props.onClick || this.handleOnClick}
                block={this.props.block}
                style="default"
                iconClass="fa fa-chevron-left" >
                Regresar
        </Button>
        );
    }
}

/**Un botón con forma de un icono */
export class Link extends React.PureComponent<ButtonProps, {}>
{
    static colors = {
        default: undefined,
        primary: "#26b",
        success: "#1a5",
        info: "#26b",
        warning: "#dd2",
        danger: "#d22",
        link: "#26b",
    }
    buttonFactory = (onClick: (() => void) | undefined, cargando: boolean) => {
        const style = { margin: 3, cursor: "pointer", color: Link.colors[this.props.style || "default"] };
        return (
            cargando ?
                <a style={{ margin: 3 }}>
                    <i className="fa fa-circle-o-notch fa-spin" />
                </a>
                :
                <a
                    onClick={onClick}
                    title={this.props.tooltip}
                    style={style}
                >
                    {this.props.iconClass && <i className={this.props.iconClass} />} {this.props.children}
                </a>
        );
    }

    errorFactory = (error, onClick: () => void) =>
        <a
            onClick={onClick}
            title={"Error: " + error}
            style={{ color: "red" }}
        >
            <i className="fa fa-times" />
        </a>

    render() {
        return (
            <PromiseAction
                buttonFactory={onClick => this.buttonFactory(onClick, false)}
                cargandoComponent={this.buttonFactory(undefined, true)}
                errorFactory={this.errorFactory}
                onClick={this.props.onClick}
            />
        );
    }
}

export class IconButton extends React.PureComponent<ButtonProps, {}>{
    constructor(props) {
        super(props);
        console.warn("IconButton esta obsoleto, utilice la clase Link");
    }

    render() {
        return React.createElement(Link, this.props, this.props.children);
    }
}

/**Boton para borrar un elemento de un listado. Tiene el icono del bote de basura */
export class BorrarIcon extends React.PureComponent<{ onClick: () => void }, {}>
{
    render() {
        return <Link onClick={this.props.onClick} iconClass="fa fa-trash-o" tooltip="Eliminar" style="danger" />
    }
}

/**Boton para editar un elemento de un listado. Tiene el icono del lapiz */
export class EditButton extends React.PureComponent<{ onClick: () => void }, {}>
{
    render() {
        return <Link onClick={this.props.onClick} iconClass="fa fa-pencil" tooltip="Editar" style="primary" />
    }
}

/**Botón para ver un elemento de un listado. Tiene el icono del ojo */
export class VerButton extends React.PureComponent<{ onClick: () => void }> {
    render() {
        return <Link onClick={this.props.onClick} iconClass="fa fa-eye" tooltip="Ver" style="primary" />
    }
}

/**Propiedades de los botones por default de una fila */
export interface RowButtonProps {
    onEdit?: () => void;
    onDelete?: () => void;
}

/**Botones de una fila de un catalogo, que son el de editar y el de borrar */
export class RowButtons extends React.PureComponent<RowButtonProps, {}>  {
    render() {
        return (
            <span>
                {this.props.onEdit && <EditButton onClick={this.props.onEdit} />}
                {this.props.onDelete && <BorrarIcon onClick={this.props.onDelete} />}
            </span>
        );
    }
}

