import chartjs = require("chart.js");
import React = require("react");

export type ChartType = "pie" | "doughnut" | "bar" | "line";

export interface ChartSeries {

    values: number[];
    title?: string;
    /**Arreglo de colores css */
    colors?: string[] | string;
}

export interface ChartData {
    type: ChartType;
    series: ChartSeries[];
    labels: string[];
}

type RGB = [number, number, number];

const colors: RGB[] = [
    [54, 162, 235],
    [255, 99, 132],
    [255, 206, 86],
    [75, 192, 192],
    [153, 102, 255],
    [255, 159, 64]
];

/**Obtiene los colores de una gráfica */
function getChartColors(count: number): RGB[] {
    const ret: RGB[] = [];
    for (var i = 0; i < count; i++) {
        ret.push(colors[i % colors.length]);
    }
    return ret;
}

function toCssColor(alpha: number, color: RGB): string {
    return `rgba(${color[0]}, ${color[1]}, ${color[2]}, ${alpha})`;
}

function getPieChartData(data: ChartData) {
    const colorPorSerie = data.type == "bar" || data.type == "line";
    const cantColores = colorPorSerie ? data.series.length : data.series.map(x => x.values.length).reduce((a, b) => Math.max(a, b), 0);

    const colors = getChartColors(cantColores);
    const backgroundColors = colors.map(x => toCssColor(0.5, x));
    const foreColors = colors.map(x => toCssColor(1, x));

    const type = data.type == "bar" ? "bar" : data.type;

    if (data.series.map(x => x.values.length).filter(x => x != data.labels.length).length > 0) {
        throw "La cantidad de labels debe de ser igual a la cantidad de valores en todas las series";
    }
    const options = type == "bar" || type == "line" ?
        {
            legend: {
                display: false
            },
            scales: {
                yAxes: [{
                    display: true,
                    ticks: {
                        suggestedMin: 0,    // minimum will be 0, unless there is a lower value.
                        // OR //
                        beginAtZero: true   // minimum value will be 0.
                    }
                }]
            }
        } : undefined;

    const getSeriesData = (serie: ChartSeries, index: number) => {
        return {
            data: serie.values,
            backgroundColor: serie.colors || (colorPorSerie ? backgroundColors[index] : backgroundColors),
            hoverBackgroundColor: serie.colors || (colorPorSerie ? backgroundColors[index] : backgroundColors),
            pointHoverBackgroundColor: backgroundColors[0],
            pointHoverBorderColor: foreColors[0],
            pointBackgroundColor: backgroundColors[0]
        };
    }

    return (
        {
            responsive: true,
            type: type,
            data: {
                labels: data.labels,
                datasets: data.series.map(getSeriesData)
            },
            options: options
        });
}

export interface ChartComponentProps {
    value: ChartData;
    style?: React.CSSProperties;
}

/**Una gráfica que soporta varias series y varios tipos de gráficas*/
export class Chart extends React.PureComponent<ChartComponentProps, {}> {
    private onCanvasInit = (canvas: HTMLCanvasElement) => {
        this.canvas = canvas;
        if (canvas)
            this.updateChartData(this.props);
        else
            this.destroyChart();
    }

    private destroyChart() {
        if (this.chart != null) {
            this.chart.destroy();
        }
        this.chart = null;
    }

    private updateChartData(props: ChartComponentProps) {
        const chartData = getPieChartData(props.value);
        this.destroyChart();
        this.chart = new chartjs(this.canvas, chartData);
    }

    canvas: HTMLCanvasElement;
    chart: any;
    componentWillUpdate(props: ChartComponentProps) {
        this.updateChartData(props);
    }

    render() {
        return (
            <div style={this.props.style}>
                <canvas ref={this.onCanvasInit} >
                </canvas>
            </div>
        );
    }
}

export interface SimpleChartValue {
    label: string;
    value: number;
}

export interface SimpleChartProps {
    /**Tipo de la gráfica. Por default es 'doughnut' */
    type?: ChartType;
    /**Valores de la gráfica */
    values: SimpleChartValue[];
    /**Arreglo de colores */
    colors?: string[];
}

/**Gráfica que soporta sólo una serie de información */
export class SimpleChart extends React.PureComponent<SimpleChartProps, {}> {
    render() {
        const data: ChartData = {
            type: this.props.type || "doughnut",
            labels: this.props.values.map(x => x.label),
            series: [
                {
                    values: this.props.values.map(x => x.value),
                    colors: this.props.colors
                }
            ]
        };

        return (
            <Chart value={data} />
        )
    }
}

export interface HistogramChartProps {
    values: (string | number)[];
    sort?: (a: string, b: string) => number;
}
/**Una gráfica que devuelve la cantidad de elemento repetidos */
export class HistogramChart extends React.PureComponent<HistogramChartProps, {}> {
    render() {
        const data = new Map<number | string, number>();

        //Acumulamos las frecuencias:
        for (const id of this.props.values) {
            data.set(id, (data.get(id) || 0) + 1);
        }

        const values = Array.from(data.keys()).sort(this.props.sort).map(key => ({ label: "" + key, value: data.get(key)! }));

        return (
            <SimpleChart type="bar" values={values} />
        )
    }
}