import React = require("react");
import Field = require("./formField");
import interfaces = require("../interfaces");
import NumberInput = require("react-number-input");
import { async } from "../async";
import { FastInput } from "./fastInput";
import { FormatCurrency } from "../format";
import { Spinner } from "../cargando";
import { SlowInput } from "./slowInput";
interface TextInputProps extends interfaces.FormEditProps<string | undefined | null, string> {
    type: "text" | "number" | "password";
    addon?: JSX.Element | string;
    placeholder?: string;
    multiline?: boolean;
    /**True para deshabilitar el uso del fastInput, el cual filtra los eventos de onChange para mejorar el rendimiento pero queda descontrolado el input */
    disableFastInput?: boolean;
}

class TextIcon extends React.PureComponent<TextInputProps, {}> {
    render() {
        //Quitamos los props que no le pertenecen al FastInput
        //ver https://github.com/facebook/react/issues/7266
        const { addon, error, isReadOnly, value, label, disableFastInput, ...props } = this.props;

        const input = disableFastInput ? <SlowInput
            {...props}
            className="form-control"
            value={value || ""}
            disabled={isReadOnly}
        /> :
            <FastInput
                {...props}
                className="form-control"
                value={value || ""}
                disabled={isReadOnly}
                type={props.type || "text"}
            />;
        return (
            addon ?
                <div className="input-group">
                    <span className="input-group-addon">
                        {addon}
                    </span>
                    {input}
                </div> : input
        )
    }
}

class TextInput extends React.PureComponent<TextInputProps, {}> {
    render() {
        return (
            <Field.FormField label={this.props.label} error={this.props.error}>
                <TextIcon  {... this.props} />
            </Field.FormField>
        );
    }
}

/**Un campo de texto */
export class Text extends React.PureComponent<interfaces.AsyncFormEditProps<string | undefined | null, string> & { addon?: JSX.Element | string, placeholder?: string, multiline?: boolean, disableFastInput?: boolean }, {}>
{
    render() {
        return (
            async(this.props.value, x =>
                <TextInput
                    {... { ... this.props, type: "text", value: (x || "") }}
                />, {
                    cargando: <TextInput
                        isReadOnly={this.props.isReadOnly}
                        onChange={this.props.onChange}
                        value={""}
                        placeholder="Cargando..."
                        multiline={this.props.multiline}
                        type="text"
                        addon={<Spinner />}
                        disableFastInput
                    />
                })
        );
    }
}

export interface NumberInputProps extends interfaces.FormEditProps<number | undefined | null, number> {
    format?: "decimal" | "int" | "currency" | "percentage" | "int-2-digit";
    placeholder?: string;
    min?: number;
    max?: number;
}
/**Un campo numérico */
export class Number extends React.PureComponent<NumberInputProps, {}>
{
    render() {
        const format =
            this.props.format == "decimal" ? "0,0[.]00" :
                this.props.format == "int" ? "0" :
                    this.props.format == "currency" ? "$0,0.00" :
                        this.props.format == "percentage" ? "0%" :
                            this.props.format == "int-2-digit" ? "00" :
                                this.props.format == undefined ? "0,0[.][00]" :
                                    this.props.format;

        const text = this.props.format == "currency" ? FormatCurrency(this.props.value) : (this.props.value + "");
        const textProps = { ... this.props, value: text, onChange: undefined, isReadOnly: true };

        const readOnly = <Text {...textProps } />;
        const editor =
            <Field.FormField label={this.props.label} error={this.props.error}>
                <NumberInput.default
                    className="form-control"
                    value={this.props.value}
                    onChange={this.props.onChange}
                    format={format}
                    placeholder={this.props.placeholder}
                    min={this.props.min}
                    max={this.props.max}
                />
            </Field.FormField>;

        return this.props.isReadOnly ? readOnly : editor;
    }
}

/**Un campo de contraseña */
export class Password extends React.PureComponent<interfaces.FormEditProps<string | undefined | null, string> & { addon?: JSX.Element | string }, {}>
{
    render() {
        return (
            <TextInput
                type="password"
                error={this.props.error}
                isReadOnly={this.props.isReadOnly}
                label={this.props.label}
                onChange={this.props.onChange}
                value={this.props.value}
                addon={this.props.addon}
            />
        );
    }
}