import React = require("react");
import {
    CellData, GridRowExEx, GridColumnEx, GridRow,
    filtroKind, BaseFiltro, FiltroRangoNumerico, FiltroPorValorUnico, FiltroRangoFechas, FiltroPorBusqueda, Filtro,
    FiltroColumna, DataTypes
} from "./tipos";
import { obtenerTiposColumnas, compare, valoresUnicos } from "./logic";
import { handleFormChange } from "../form";
import { Number, Text, DateTime } from "../fields/fields";
import { MultiSelect } from "../selector";
import { Link } from "../botones";
import { search } from "../search";
import { formatCellData } from "./format";

/**Obtiene un arreglo con los tipos de filtros adecuados segun la información, el arreglo tiene un elemento por cada celda */
export function obtenerTiposFiltro(tiposColumnas: DataTypes[]): (filtroKind | null)[] {
    const ret = tiposColumnas.map(x => (
        x == "string" ? "valorUnico" :
            x == "boolean" ? "valorUnico" :
                x == "null" ? null :
                    x == "Date" ? "rangoFecha" :
                        x == "number" ? "rangoNumerico" : null
    ) as filtroKind)

    return ret;
}

/**Agrega un nuevo filtro a un filtro columna, devuelve el nuevo arreglo de filtro columna */
export function crearFiltro(tipoFiltro: filtroKind): Filtro {
    switch (tipoFiltro) {
        case "rangoFecha":
            return { kind: "rangoFecha" };
        case "rangoNumerico":
            return { kind: "rangoNumerico" };
        case "valorUnico":
            return { kind: "valorUnico", valores: [] };
        case "busqueda":
            return { kind: "busqueda", texto: "" };
    }
}

interface FiltroProps<T extends BaseFiltro> {
    value: T;
    onChange: (value: T) => void;
    filas: PromiseLike<GridRow[]> | GridRow[];
    columnas: GridColumnEx[];
    columna: number;
    onQuitar: () => void;
}

class FiltroRangoNumericoComponent extends React.PureComponent<FiltroProps<FiltroRangoNumerico>, {}> {
    private handleChange = handleFormChange(() => this.props);
    render() {
        const x = this.props.value;
        const change = this.handleChange;
        const titulo = this.props.columnas[this.props.columna].titulo;
        return (
            <div>
                <div className="col-md-6">
                    <Number value={x.inicio} label={<span><Link iconClass="fa fa-times" onClick={this.props.onQuitar} /> <label>{titulo + " inicio"}</label></span>} onChange={change("inicio")} />
                </div>
                <div className="col-md-6">
                    <Number value={x.final} label={titulo + " final"} onChange={change("final")} />
                </div>
            </div>
        );
    }
}

class FiltroRangoFechasComponent extends React.PureComponent<FiltroProps<FiltroRangoFechas>, {}> {
    private handleChange = handleFormChange(() => this.props);
    render() {
        const x = this.props.value;
        const change = this.handleChange;
        const titulo = this.props.columnas[this.props.columna].titulo;
        return (
            <div>
                <div className="col-md-6">
                    <DateTime value={x.inicio} label={<span><Link iconClass="fa fa-times" onClick={this.props.onQuitar} /> <label>{titulo + " inicio"}</label></span>} onChange={change("inicio")} />
                </div>
                <div className="col-md-6">
                    <DateTime value={x.final} label={titulo + " final"} onChange={change("final")} />
                </div>
            </div>
        );
    }
}

class FiltroBusquedaTextoComponent extends React.PureComponent<FiltroProps<FiltroPorBusqueda>, {}> {
    private handleChange = (value: string) => {
        if (this.props.onChange) {
            this.props.onChange({ ... this.props.value, texto: value });
        }
    }

    render() {
        const titulo = this.props.columnas[this.props.columna].titulo;
        const label = <span><Link iconClass="fa fa-times" onClick={this.props.onQuitar} /> <label>{titulo}</label></span>;
        return (
            <div>
                <Text value={this.props.value.texto} onChange={this.handleChange} label={label} />
            </div>
        );
    }
}

class FiltroInvalido extends React.PureComponent<FiltroProps<any>> {
    render() {
        return <a style={{ color: "red", cursor: "pointer" }} onClick={this.props.onQuitar} >Filtro inválido, click aqui para quitar el filtro</a>
    }
}

class FiltroValorUnicoComponent extends React.PureComponent<FiltroProps<FiltroPorValorUnico>, {}> {
    private async unicos() {
        return valoresUnicos(await this.props.filas, this.props.columna);
    }

    private async items() {
        return (await this.unicos()).map(x => ({ value: x, label: formatCellData(x) }));
    }

    private handleChange = handleFormChange(() => this.props);
    render() {
        const x = this.props.value;
        const change = this.handleChange;
        const items = this.items();
        const titulo = this.props.columnas[this.props.columna].titulo;
        const label = <span><Link iconClass="fa fa-times" onClick={this.props.onQuitar} /> <label>{titulo}</label></span>;
        return (
            <div>
                <MultiSelect value={x.valores} onChange={change("valores")} items={items} label={label} />
            </div>
        )
    }
}



class FiltroComponent extends React.PureComponent<FiltroProps<Filtro> & { onQuitar?: () => void }, {}> {
    render() {
        //El filtro puede ser invalido cuando el usuario seleccione ciertos parametros del filtro, estos se guarden en el cache, y en
        //versiones posteriores esas columnas ya no existan en la información
        const filtroInvalido = this.props.columna >= this.props.columnas.length;
        const kind = this.props.value.kind;
        const tipo: React.ComponentClass<FiltroProps<Filtro>> =
            filtroInvalido ? FiltroInvalido :
                kind == "rangoFecha" ? FiltroRangoFechasComponent :
                    kind == "rangoNumerico" ? FiltroRangoNumericoComponent :
                        kind == "valorUnico" ? FiltroValorUnicoComponent :
                            kind == "busqueda" ? FiltroBusquedaTextoComponent :
                                null as never;
        const comp = React.createElement(tipo, { ... (this.props as any) });
        return comp;
    }
}

export interface FiltroListProps {
    value: FiltroColumna[];
    onChange?: (x: FiltroColumna[]) => void;
    filas: GridRow[] | PromiseLike<GridRow[]>;
    columnas: GridColumnEx[];
}

export class FiltroListComponent extends React.PureComponent<FiltroListProps, {}> {
    private handleOnChange = (value: Filtro, indice: number) => {
        if (this.props.onChange) {
            const nuevo = this.props.value.map((x, i) => i == indice ? { ...x, datos: value } : x);
            this.props.onChange(nuevo);
        }
    };

    private handleQuitar = (indice: number) => {
        if (this.props.onChange) {
            this.props.onChange(this.props.value.filter((x, i) => i != indice));
        }
    }

    render() {
        return (
            this.props.value.length > 0 ?
                <div>
                    <div className="row">
                        {
                            this.props.value.map((x, i) =>
                                <div className="col-md-4">
                                    <FiltroComponent
                                        filas={this.props.filas}
                                        columnas={this.props.columnas}
                                        columna={x.indice}
                                        value={x.datos}
                                        onChange={valor => this.handleOnChange(valor, i)}
                                        onQuitar={() => this.handleQuitar(i)}
                                    />
                                </div>
                            )
                        }
                    </div>
                </div> : null
        );
    }
}