import React = require("react");
import {
    toGridRowExEx, toGridColumnEx, obtenerTiposColumnas, compare, ObtenerDatosCsv,
    filtrarOrdenar, getGridHeader, getGridRow
} from "./logic";
import {
    GridRow, GridRowExEx, OrdenGrid, CellData, GridColumn, GridColumnEx,
    FiltroColumna, filtroKind
} from "./tipos";
import { PaginatedTable } from "../pagination";
import { FiltroListComponent, crearFiltro, obtenerTiposFiltro } from "./filtros";
import { async } from "../async";
import { Text } from "../fields/text";
import { Cookies } from "../cookies";
import { Link, Button } from "../botones";
import { download } from "../downloadjs";
import { Base64ToBlob, stringToBlob } from "../files"
import { toCSV } from "./csv";
import memoize = require("memoizeref");
import { MakeUncontrolledCookies } from "../HOC";
import { PaginatedTableState } from "../table";
/**
* Propiedades del grid
*/
interface SyncGridProps {
    /**Definición de las columans del grid */
    columns: GridColumn[];
    /**Datos del grid */
    rows: GridRow[] | PromiseLike<GridRow[]>;

    /**Objeto que se va a pasar a los metodos de generación de la fila visual de los rows. Utilice este objeto para cambiar la apariencia de las filas
     * sin necesidad de recalcular los filtros, paginación u ordenamientos del grid
     */
    visualProps?: any;

    /**Valores del filtro y del orden.*/
    gridState?: GridState;
    /**Se llama al modificar el filtro orden.*/
    gridStateChange?: (value: GridState) => void;

    /**True para ocultar el cuadro de búsqueda del grid */
    hideBusqueda?: boolean;
}

export interface GridState {
    orden?: OrdenGrid;
    filtros: FiltroColumna[]
    busqueda: string;
    pagination?: PaginatedTableState
}

/**Obtiene la cabecera de un grid */
function getTableHeader(columns: GridColumnEx[]) {
    const cells = columns.map(x => x.header)
    return (
        <tr>
            {...cells}
        </tr>);
}

/**
 * El grid es un control similar a la tabla paginada, pero que toma los datos en lugar de las filas de la tabla, es un
 * control menos flexible pero que ofrece posibilidad de ordenar los datos por columna, filtrar los datos de manera dinámina
 * y de exportar a excel o CSV ya que tiene acceso a los datos sin procesar
 */
class SyncGrid extends React.PureComponent<SyncGridProps> {
    private get gridState(): GridState {
        return this.props.gridState || { filtros: [], busqueda: "" };
    }

    private setFiltroOrden(value: Partial<GridState>) {
        if (this.props.gridStateChange) {
            const newValue = { ... this.gridState, ...value };
            this.props.gridStateChange(newValue);
        }
    }

    /**
     * Alterna el ordenamiento de la columna especificada
     */
    private handleOnOrdenar = (index: number) => {
        const orden = this.gridState.orden;
        if (!orden || orden.columna != index) {
            this.setFiltroOrden({ orden: { columna: index, orden: "asc" } });
        } else {
            this.setFiltroOrden({ orden: orden.orden == "asc" ? { columna: index, orden: "desc" } : undefined });
        }
    }

    /**
     * Agrega o quita el filtro de cierta columna
     */
    private handleOnFiltrar = (index: number, tipoFiltro: filtroKind) => {
        const estaFiltrado = this.gridState.filtros.map(x => x.indice).indexOf(index) != -1

        const nuevoFiltro: FiltroColumna = {
            datos: crearFiltro(tipoFiltro),
            indice: index
        };

        const nuevoArr = estaFiltrado ?
            this.gridState.filtros.filter(x => x.indice != index) :  //Quitar el filtro
            [...this.gridState.filtros, nuevoFiltro];                //Poner el filtro

        this.setFiltroOrden({ filtros: nuevoArr });
    }

    private handlePaginationChange = (value: PaginatedTableState) => {
        this.setFiltroOrden({ pagination: value });
    }

    /**
     * Descarga la información que esta viendo el usuario en un CSV
     */
    private handleDownload = async () => {
        const dataFiltered = this.filtrarOrdenar(await this.props.rows, this.gridState.orden, this.gridState.filtros, this.gridState.busqueda, this.props.columns.length);
        const data = ObtenerDatosCsv(dataFiltered.map(x => toGridRowExEx(x)), this.props.columns.map(x => toGridColumnEx(x)));
        const csv = toCSV(data);
        const blob = stringToBlob(csv);
        download(blob, "Lista.csv");
    }

    private filtrarOrdenar = memoize(filtrarOrdenar);
    private getHeader = memoize(getGridHeader);
    private obtenerTiposColumnas = memoize(obtenerTiposColumnas);
    private obtenerTiposFiltro = memoize(obtenerTiposFiltro);
    private obtenerColumnasNormalizadas = memoize((cols: GridColumn[]) => cols.map(x => toGridColumnEx(x)));

    private async getDataFiltered() {
        return this.filtrarOrdenar(await this.props.rows, this.gridState.orden, this.gridState.filtros, this.gridState.busqueda, this.props.columns.length);
    }
    private async getRows() {
        const data = (await this.getDataFiltered());
        const rows = data.map((row, i) => getGridRow(row, i, () => this.props.visualProps));
        return rows;
    }
    private async getTiposFiltros() {
        const tipos = this.obtenerTiposColumnas(await this.props.rows, this.props.columns.length);
        const filtros = this.obtenerTiposFiltro(tipos);
        return filtros;
    }

    private getGridHeader(columnas: GridColumnEx[]) {
        const filtrosProm = this.getTiposFiltros();
        const sinFiltros = columnas.map(x => null);
        return async(
            filtrosProm,
            filtros => this.getHeader(columnas, this.gridState.orden, filtros, this.gridState.filtros, this.handleOnOrdenar, this.handleOnFiltrar),
            { cargando: this.getHeader(columnas, this.gridState.orden, sinFiltros, this.gridState.filtros, this.handleOnOrdenar, this.handleOnFiltrar) });
    }

    render() {
        const orden = this.gridState.orden;
        const dataFiltered = this.getDataFiltered();
        const filtros = this.getTiposFiltros();
        const columnas = this.obtenerColumnasNormalizadas(this.props.columns);
        const header = this.getGridHeader(columnas);

        const rows = this.getRows();
        return (
            <div>
                <FiltroListComponent
                    filas={this.props.rows}
                    columnas={columnas}
                    onChange={x => this.setFiltroOrden({ filtros: x })}
                    value={this.gridState.filtros}
                />
                {
                    !this.props.hideBusqueda &&
                    <div style={{ display: "flex", flexDirection: "row" }}>
                        <div style={{ flex: "1 1 auto" }}>
                            <Text value={this.gridState.busqueda}
                                placeholder="Búsqueda" onChange={x => this.setFiltroOrden({ busqueda: x })} addon={<i className="fa fa-search" />} />
                        </div>
                        <div style={{ flex: "0 1 auto" }}>
                            <Button tooltip="Descargar CSV" style="success" iconClass="fa fa-download" onClick={this.handleDownload} />
                        </div>
                    </div>
                }
                <PaginatedTable header={header} items={rows} value={this.gridState.pagination} onChange={this.handlePaginationChange} />
            </div>
        )
    }
}

export interface GridProps {
    /**Definición de las columans del grid */
    columns: PromiseLike<GridColumn[]> | GridColumn[];
    /**Datos del grid */
    rows: PromiseLike<GridRow[]> | GridRow[];
    /**Objeto que se va a pasar a los metodos de generación de la fila visual de los rows. Utilice este objeto para cambiar la apariencia de las filas
     * sin necesidad de recalcular los filtros, paginación u ordenamientos del grid
     */
    visualProps?: any;

    /**Valores del filtro y del orden.*/
    gridState?: GridState;
    /**Se llama al modificar el filtro orden.*/
    onGridStateChange?: (value: GridState) => void;

    /**True para ocultar el cuadro de búsqueda */
    hideBusqueda?: boolean;
}

export class GridControlled extends React.PureComponent<GridProps> {
    render() {
        return async(Promise.all([this.props.columns]), all =>
            <SyncGrid
                columns={all[0]}
                rows={this.props.rows}
                visualProps={this.props.visualProps}
                gridState={this.props.gridState}
                gridStateChange={this.props.onGridStateChange}
                hideBusqueda={this.props.hideBusqueda}
            />)
    }
}

//Habilitamos el uso descontrolado de filtroOrden:
export const Grid = MakeUncontrolledCookies<GridProps>("gridState", "onGridStateChange")(GridControlled);
