//Contiene la mayoria de la logica para el grid
//Todas estas funciones son puras

import React = require("react");
import formats = require("../format");
import {
    CellData, GridCell, GridCellEx, GridCellExEx, GridDataFormat, GridRow, GridRowEx, GridRowExEx, GridColumn, GridColumnEx, ColumnaProps,
    OrdenGrid, FiltroColumna, Filtro, DataTypes, filtroKind
} from "./tipos";
import { formatCellData, getCellDataType } from "./format";
import { search } from "../search";
import { GridCellComponent, GridHeaderComponent } from "./components";



function isGridColumnEx(x: GridColumn): x is GridColumnEx {
    const y = x as GridColumnEx;
    return (typeof y.header == "function");
}

/**Convierte a GridColumnEx */
export function toGridColumnEx(x: GridColumn): GridColumnEx {
    if (isGridColumnEx(x)) {
        return x;
    } else if (typeof x === "string") {
        return {
            titulo: x,
            header: (props) => <GridHeaderComponent {...props} titulo={x} />
        }
    } else if (typeof x === "function") {
        return { header: x, titulo: "" };
    }
    else {
        return { header: (props) => x, titulo: "" }
    }
}



/**
 * Devuelve si un GridCell es un CellData
 * @param x 
 */
function isCellData(x: GridCell): x is CellData {
    const ret =
        (typeof x === "string") ||
        (typeof x === "number") ||
        (typeof x === "boolean") ||
        (x == null) ||
        (x instanceof Date);
    return ret;
}

/**Devuelve si un GridCell es un GridCellEx */
function isGridCellExEx(x: GridCell): x is GridCellExEx {
    if (!x) return false;
    const y = x as GridCellExEx;
    return isCellData(y.data) && React.isValidElement(y.view);
}

function isGridCellEx(x: GridCell): x is GridCellEx {
    if (!x) return false;
    const y = x as GridCellEx;
    return isCellData(y.data) && (typeof y.format == "string");
}



/**Normaliza un GridCell a un GridCellEx */
function toGridCellEx(x: GridCell, index: number): GridCellExEx {
    if (isGridCellExEx(x)) {
        return x;
    } else if (isGridCellEx(x)) {
        return { data: x.data, view: <GridCellComponent data={x.data} format={x.format} /> }
    } else if (isCellData(x)) {
        return { data: x, view: <GridCellComponent data={x} /> };
    } else {
        return { data: undefined, view: x };
    }
}

/**Determina si un GridRowEx o ExEx es un GridRowExEx */
function isGridRowExEx(x: GridRowEx | GridRowExEx): x is GridRowExEx {
    const y = x as GridRowExEx;
    return !!y.visual;
}


/**Normaliza un GridRow a un GridRowExEx */
export function toGridRowExEx(x: GridRow): GridRowExEx {
    if (Array.isArray(x)) {
        const y: GridRowEx = { data: x };
        return toGridRowExEx(y);
    } else if (isGridRowExEx(x)) {
        return x;
    } else {
        const data = x.data.map(y => getCellData(y));
        const ret: GridRowExEx = {
            visual: () => x.data.map((x, i) => toGridCellEx(x, i)),
            data: (i) => data[i],
            props: x.props
        }

        return ret;
    }
}

/**Obtiene un arreglo que por cada columna, devuelve el tipo de dato que mas existe en esa columna */
export function obtenerTiposColumnas(data: GridRow[], columnCount: number): DataTypes[] {
    //Tratamos de tomar una muestra de este tamaño:
    const sampleSize = 100;
    const step = Math.round(Math.max(data.length / sampleSize, 1));
    const conteoTipos: { [columna: number]: { [tipo: string]: number } } = {};

    //Armamos el arreglo de conteo de tipos, tomando una muestra de las filas
    for (let y = 0; y < data.length; y += step) {
        const row = getRowCellData(data[y], columnCount);

        for (let x = 0; x < row.length; x++) {
            const cellData = row[x];
            const columna = conteoTipos[x] = conteoTipos[x] || {};
            const tipo = getCellDataType(cellData);
            columna[tipo] = (columna[tipo] || 0) + 1;
        }
    }

    //Cantidad de columnas:
    const cantColumnas = 1 + Object.keys(conteoTipos).map(x => Number.parseInt(x)).reduce((a, b) => Math.max(a, b), 0);

    //Por cada columna, el tipo que tiene mas cantidad de elementos o null si esa columna no tenia
    const tiposColumnas: DataTypes[] = [];
    for (let i = 0; i < cantColumnas; i++) {
        const columna = conteoTipos[i] || {};
        //Arreglo de {tipo, cantidad}
        const conteosPorTipo = Object.keys(columna).map(x => x as DataTypes).map(x => ({ tipo: x, conteo: columna[x] }));
        //El tipo con el conteo maximo
        const max = conteosPorTipo.sort((a, b) => b.conteo - a.conteo)[0];
        const vacio = max == null;

        tiposColumnas[i] = vacio ? "null" : max.tipo;
    }
    return tiposColumnas;
}

export function compare(a, b) {
    return (
        a == b ? 0 :
            a == null ? 1 :
                b == null ? -1 :
                    a > b ? 1 :
                        a < b ? -1 : 0);
}

function enumerateThunk<T>(dataThunk: (index: number) => T, count: number) {
    const ret = new Array<T>(count);
    for (let i = 0; i < count; i++) {
        ret[i] = dataThunk(i);
    }
    return ret;
}

/**Obtiene los datos que se deben de descargar en el CSV de un grid */
export function ObtenerDatosCsv(rows: GridRowExEx[], columns: GridColumnEx[]) {
    const types = obtenerTiposColumnas(rows, columns.length);
    const cabecera = columns.map(x => x.titulo).filter((x, i) => types[i] && types[i] != "null");


    const datos = rows.map(row => enumerateThunk(row.data, columns.length).filter((x, i) => types[i] && types[i] != "null"));

    return [cabecera, ...datos];
}


function busquedaCadena(busqueda: string, elemento: string) {
    return elemento.toLocaleLowerCase().indexOf(busqueda.toLocaleLowerCase()) != -1;
}

function busquedaRango(min: number | undefined, max: number | undefined, valor: number | undefined) {
    if (min == null && max == null) return true;
    if (valor == null) return false;
    return (!min || valor >= min) && (!max || valor <= max);
}

function busquedaRangoFecha(min: Date | undefined, max: Date | undefined, valor: Date | undefined) {
    return busquedaRango(min && min.valueOf(), max && max.valueOf(), valor && valor.valueOf());
}

function busquedaValoresUnicos(valores: CellData[], valor: CellData) {
    if (valores.length == 0) return true;
    return valores.indexOf(valor) != -1;
}

function aplicarFiltro(filtro: Filtro, datos: CellData): boolean {
    switch (filtro.kind) {
        case "busqueda":
            return busquedaCadena(filtro.texto, formatCellData(datos));
        case "rangoNumerico":
            return busquedaRango(filtro.inicio, filtro.final, datos as number);
        case "rangoFecha":
            return busquedaRangoFecha(filtro.inicio || undefined, filtro.final || undefined, datos as Date);
        case "valorUnico":
            return busquedaValoresUnicos(filtro.valores, datos);
    }
}

/**Obtiene los datos de celda de un GridCell */
function getCellData(x: GridCell): CellData {
    if (isGridCellExEx(x)) {
        return x.data;
    } else if (isGridCellEx(x)) {
        return x.data;
    } else if (isCellData(x)) {
        return x;
    } else {
        return undefined;
    }
}

/**Obtiene los datos de celda de un GridRow */
function getRowCellData(row: GridRow, cellCount: number): CellData[] {
    if (Array.isArray(row)) {
        return row.map(x => getCellData(x));
    } else if (isGridRowExEx(row)) {
        return enumerateThunk(row.data, cellCount);
    } else {
        return row.data.map(x => getCellData(x));
    }
}

/**Obtiene los datos de celda de un GridRow en cierta columna */
export function getRowCellDataAt(row: GridRow, index: number): CellData {
    if (Array.isArray(row)) {
        return getCellData(row[index]);
    } else if (isGridRowExEx(row)) {
        return row.data(index);
    } else {
        return getCellData(row.data[index]);
    }
}

/**Obtiene los valroes unicos de una columna */
export function valoresUnicos(datos: GridRow[], columna: number) {
    const valores = datos.map(fila => getRowCellDataAt(fila, columna));
    const set = new Set(valores);
    return Array.from(set).sort((a, b) => compare(a, b));
}

export function aplicarFiltroBusqueda<T>(datos: T[], getKey: (item: T) => GridRow, filtro: string, columnCount: number) {
    if (!filtro) return datos;
    const toStr = (row: GridRow) => getRowCellData(row, columnCount).map(x => formatCellData(x)).reduce((a, b) => a + " " + b);
    return datos.filter(x => search(filtro, toStr(getKey(x))));
}

/**Aplica un filtro a la información */
export function aplicarFiltroColumna<T>(datos: T[], getKey: (item: T) => GridRow, filtro: FiltroColumna) {
    return datos.filter(x => aplicarFiltro(filtro.datos, getRowCellDataAt(getKey(x), filtro.indice)));
}

/**Devuelve los datos en orden de los datos que se van a mostrar en un grid */
export function filtrarOrdenar(data: GridRow[], orden: OrdenGrid | undefined, filtros: FiltroColumna[], busqueda: string, columnCount: number) {
    let ret = data;
    //Aplicamos el orden:
    if (orden) {
        const sortFunc = (a: CellData, b: CellData) => (compare(a, b)) * (orden.orden == "asc" ? 1 : -1);
        const getColumnData = (x: GridRow) => getRowCellDataAt(x, orden.columna);
        const rowSortFunc = (a: GridRow, b: GridRow) => sortFunc(getColumnData(a), getColumnData(b));

        ret = [...ret].sort(rowSortFunc);
    }
    ret = filtrarItems(ret, x => x, filtros, busqueda, columnCount);
    return ret;
}

/**
 * Filtra una colección de elementos dado el filtro de un grid
 * @param data 
 * @param getKey 
 * @param filtros 
 * @param busqueda 
 * @param columnCount 
 */
export function filtrarItems<T>(data: T[], getKey: (item: T) => GridRow, filtros: FiltroColumna[], busqueda: string, columnCount: number): T[] {
    let ret = data;
    ret = filtros.reduce((a, b) => aplicarFiltroColumna(a, getKey, b), ret);
    ret = aplicarFiltroBusqueda(ret, getKey, busqueda, columnCount);
    return ret;
}

/**
 * Obtiene el elemento tr que sera la cabecera de la tabla
 * @param columnas Columnas del grid
 * @param orden Estado del ordenamiento del grid
 * @param tiposFiltros Tipos de los filtros por columna
 * @param stateFiltros Estado actual de los filtros
 * @param handleOnOrdenar Funcion que toma el indice de la columna y que se ejecuta cuando el usuario quiera ordenar por esa columna
 * @param handleOnFiltrar Funcion que toma el indice de la columna y el tipo de filtro y quita o agrega el filtro para esa columna
 */
export function getGridHeader(
    columnas: GridColumnEx[],
    orden: OrdenGrid | undefined,
    tiposFiltros: (filtroKind | null)[],
    stateFiltros: FiltroColumna[],
    handleOnOrdenar: (index: number) => void,
    handleOnFiltrar: (index: number, tipoFiltro: filtroKind) => void
) {
    const headerCells = columnas
        .map((x, i) =>
            x.header({
                onOrdenar: () => handleOnOrdenar(i),
                onFiltrar: () => handleOnFiltrar(i, tiposFiltros[i]!),
                orden: orden && orden.columna == i ? orden.orden : undefined,
                puedeOrdenar: !!tiposFiltros[i],
                puedeFiltrar: tiposFiltros[i] != null,
                estaFiltrado: stateFiltros.map(x => x.indice).indexOf(i) != -1
            }));
    return React.createElement("tr", {}, ...headerCells);
}

/**
 * Devuelve una función que dado el visualProps crea un tr del Grid
 * @param row Fila
 * @param visualProps Propiedades visuales de la fila 
 * @param key Key
 */
export function getGridRow(row: GridRow, key: string | number, visualPropsThunk: () => any) {
    return () => {
        const visualProps = visualPropsThunk();
        const x = toGridRowExEx(row);
        const visualR = x.visual(visualProps);
        const cells = visualR.map(cell => cell.view);
        return React.createElement("tr", { ...x.props, key: key }, ...cells);
    }
}
