import React = require("react");
import { componentToRx } from "react-rxk";
import { Grid2Column } from "../simpleGrid/column";
import { Grid2State, gridStateDefaultValues } from "../simpleGrid/gridState";
export type Items<T> = T[] | PromiseLike<T[]>;

import { GridHeaderComponent, gridColumnToColumnHeader } from "./header";
import { FiltroListComponent, gridColumnToFiltroColumn } from "./filtros";
import { GridBusquedaComponent } from "./busqueda"
import { PaginatedTable } from "../../pagination";
import { AsyncGridTable } from "./table";
import { filtrarOrdenar, FiltroOrdenColumna, gridColumnToFiltroOrdenColumn, getItemCellData } from "../simpleGrid/filtrarOrdenar";
import { search } from "../../search";
import { GridTableFiltro } from "./tableFiltro";
import { getEffectiveColumns } from "../simpleGrid/drillDowns";
import { createSelector, createSelectorCreator, defaultMemoize } from "reselect";
import { arrayToMap, mapObject, shallowEquals, promiseAllObj } from "keautils";
import swal = require("sweetalert");
import { downloadCsv } from "../simpleGrid/csv";
import { MakeUncontrolledCookies } from "../../HOC";
import { Observable } from "rxjs";
const createShallowSelector = createSelectorCreator(defaultMemoize, (a, b) => shallowEquals(a, b));

export interface Grid2Props<TItem, TShared> {
    columnas: Grid2Column<TItem, TShared>[];
    items?: Items<TItem>;
    getKey?: (item: TItem) => string | number;
    gridState?: Partial<Grid2State>;
    onGridStateChange?: (value: Grid2State) => void;

    /**True para ocultar el cuadro de búsqueda */
    hideBusqueda?: boolean;
    sharedProps?: TShared;
    onDescargarCsv?: () => void;
}

export type Props = Grid2Props<any, any>;
class GridControlled extends React.PureComponent<Props> {
    private items = (props: Props) => props.items == undefined ? new Promise<any[]>(() => {}) :  props.items;
    /**Columnas originales. Note que estas columnas pueden venir sin key */
    private originalColumns = (props: Props) => props.columnas;
    /**Asigna los keys por default */
    private columnSelector = createSelector(this.originalColumns, cols => cols.map(x => ({ ...x, key: x.key || x.title })));
    private originalColumnMapSelector = createSelector(this.columnSelector, cols => arrayToMap(cols, x => "" + x.key, x => x));

    private originalGridState = (props: Props) => props.gridState;
    private columnKeys = createSelector(this.columnSelector, x => x.map(y => y.key));
    private gridState = createSelector(this.originalGridState, this.columnKeys, gridStateDefaultValues)

    private drillDownSelector = createShallowSelector(this.gridState, x => x.drillDowns);
    private effectiveColumnMap = createShallowSelector(this.originalColumnMapSelector, this.drillDownSelector, 
        (cols, dds) => {
            console.log("Effective columns");
            return getEffectiveColumns(cols, dds);
        });
    private mapping = createSelector(this.gridState, state => state.mapping);

    private filtroOrdenColumnsSelector = createSelector(this.effectiveColumnMap, this.items,
        (cols, items) => promiseAllObj(mapObject(cols, value => gridColumnToFiltroOrdenColumn(items, value))));

    private itemsFiltroData = createSelector(this.items, this.filtroOrdenColumnsSelector,
        async (items, columnas) => {
            return getItemCellData(await items, await columnas);
        });

    /**Obtiene todas las columnas efectivas del grid en el orden original */
    private effectiveColumns = createSelector(this.effectiveColumnMap, this.mapping, (map, mapping) => mapping.map(key => map[key]));
    /**Obtiene los titulos de todas las columnas efectivas, en el orden original */
    private columnNames = createSelector(this.effectiveColumnMap, cols => mapObject(cols, x => x.title));

    private originalColumnCountSelector = createSelector(this.columnSelector, x => x.length);
    private filtroColumnasSelector = createSelector(this.effectiveColumnMap, this.items, (columns, items) => mapObject(columns, val => gridColumnToFiltroColumn(items, val)));

    private filtroColumnaValueSelector = createSelector(this.gridState, state => state.filtroOrden.columna);
    private onGridStateChangeSelector = (props: Props) => props.onGridStateChange;
    private filtroColumnaOnChangeSelector = createSelector(this.onGridStateChangeSelector, this.gridState, (onGridStateChange, gridState) => columna => onGridStateChange!({ ...gridState, filtroOrden: { ...gridState.filtroOrden, columna: columna } }))

    private handleExportarCSV = async () => {
        await downloadCsv(this.items(this.props), this.gridState(this.props), this.effectiveColumnMap(this.props));
    }

    render() {
        const props = this.props;
        const effectiveColumns = this.effectiveColumns(this.props);
        const originalColumnCount = this.originalColumnCountSelector(this.props);

        const filtros =
            <FiltroListComponent
                columnas={this.filtroColumnasSelector(this.props)}
                items={this.itemsFiltroData(props)}
                value={this.filtroColumnaValueSelector(this.props)}
                onChange={this.filtroColumnaOnChangeSelector(this.props)}
            />;

        const busqueda = <GridBusquedaComponent
            value={this.gridState(this.props)}
            onChange={props.onGridStateChange!}
            columnas={this.columnNames(this.props)}
            onExpotar={this.handleExportarCSV}
        />;

        const grid = <GridTableFiltro
            effectiveColumns={this.effectiveColumnMap(props)}
            getKey={props.getKey}
            gridState={this.gridState(this.props)}
            items={this.itemsFiltroData(props)}
            filtroOrdenColumnas={this.filtroOrdenColumnsSelector(props)}
            onGridStateChange={props.onGridStateChange!}
            sharedProps={props.sharedProps}
        />
        return (
            <div>
                {filtros}
                {!props.hideBusqueda && busqueda}
                {grid}
            </div>
        )
    }
}

export const Grid2 = componentToRx(MakeUncontrolledCookies<Props>("gridState", "onGridStateChange")(GridControlled));
