import React = require("react");
import { FiltroOrden, OrdenGrid, FiltroColumna, FiltroColumnaMap } from "./gridState";
import { CellData, GridCell } from "./baseTypes";
import { FiltroFunc, GridColumnFilter, ItemRowData } from "./tiposFiltros";
import { cellToText, getEffectiveString } from "./text";
import { Grid2Column } from "../simpleGrid/column";
import { search } from "../../search";
import { enumObject, ObjMap, mapObject } from "keautils";
import { } from ""
import { getAutoFiltro } from "../filtros/autoFiltro";
import { FiltroInvalidoComp } from "../components/filtroInvalido";

/**Representa una fila del grid, despues de haber obtenido los valores con el getData */

/**Obtiene el filtro efectivo de la columna, si el filtro no esta definido en la columna se usa el selector getAutoFiltro */
export async function getFiltro<TItem>(data: TItem[] | PromiseLike<TItem[]>, getData: (item: TItem) => CellData, filter: GridColumnFilter<any> | undefined) {
    const x: React.StatelessComponent<any> = () => <span />;

    if (filter) {
        return filter;
    } else {
        const dataSync = await data;
        const defaultFilter: GridColumnFilter<any> = {
            component: FiltroInvalidoComp,
            filter: () => true
        };

        const filtro = getAutoFiltro(dataSync, getData) || defaultFilter;
        return filtro;
    }
}


/**
 * Para una lista de elementos, obtiene la misma lista de elementos pero obtieniendo cada uno de los datos de las columnas con sus respectivos getCell y los datos de la busqueda,
 * esta estructura de datos se usa para acelerar las búsquedas ya que estos valores se tendran que calcular sólo cuando cambien los items, en lugar de cada vez que se realize una búsqueda
 * @param data 
 */
export function getItemCellData<T>(data: T[], columnas: ObjMap<GetDataColumna<T>>): ItemRowData<T>[] {
    console.log("Get items cell data");
    const ret = data.map(x => {
        const cell = mapObject(columnas, (value, key) => {
            const cell = value.cell(x);
            const string = cell.string;
            return {
                data: cell.data,
                string: string || cellToText(cell.data),
                style: cell.style,
                view: cell.view
            }
        });

        const textSearch = enumObject(cell).map(x => x.value.string).join(" ");
        return {
            cell,
            original: x,
            textSearch
        }
    });

    return ret;
}


/**Aplica los filtros de columnas y el orden a los datos 
 * @param searchFunc Función que se utiliza en la búsqueda por texto
*/
export function filtrarOrdenar<T>(data: ItemRowData<T>[], filtroOrden: FiltroOrden, columnas: ObjMap<FiltroOrdenColumna<T>>): ItemRowData<T>[] {
    console.log("filtrar ordenar");
    let ret = data;
    ret = filtroBusqueda(ret, filtroOrden.busqueda);
    ret = filtrarPorColumnas(ret, filtroOrden.columna, columnas);
    if (filtroOrden.orden) {
        ret = ordenar(ret, filtroOrden.orden.orden, filtroOrden.orden.columna);
    }

    return ret;
}

/**Convierte una columna del grid a una columna FiltroOrdenColumna */
export async function gridColumnToFiltroOrdenColumn<TItem>(data: TItem[] | PromiseLike<TItem[]>, value: Grid2Column<TItem, any>): Promise<FiltroOrdenColumna<TItem> & GetDataColumna<TItem>> {
    return {
        cell: value.cell,
        filtro: (await getFiltro(data,  (x)=> value.cell(x).data , value.filter)).filter,
    };
}

export function compare(a, b) {
    return (
        a == b ? 0 :
            a == null ? 1 :
                b == null ? -1 :
                    a > b ? 1 :
                        a < b ? -1 : 0);
}
/**Ordena un conjunto de datos de un grid */
function ordenar<T>(data: ItemRowData<T>[], orden: "asc" | "desc", column: string | number) {
    const getData = (x: ItemRowData<T>) => x.cell[column].data;
    const sortFunc = (a: CellData, b: CellData) => (compare(a, b)) * (orden == "asc" ? 1 : -1);
    const rowSortFunc = (a: ItemRowData<T>, b: ItemRowData<T>) => sortFunc(getData(a), getData(b));

    return [...data].sort(rowSortFunc);
}

/**Aplica los filtros por columna */
function filtrarPorColumnas<T>(data: ItemRowData<T>[], filtros: FiltroColumnaMap, columnas: ObjMap<FiltroOrdenColumna<T>>) {
    //Función que aplica un filtro:
    const aplicarFiltro = (filtro: FiltroColumna & { key: string }, data: ItemRowData<T>[]) => {
        const columna = columnas[filtro.key];
        const pasaFiltro = (x: ItemRowData<T>) => {
            const cell = x.cell[filtro.key];
            const data = cell.data;
            const string = cell.string;
            return columna.filtro(data, filtro.datos || {}, string);

        };
        return data.filter(pasaFiltro);
    };

    /**Aplicamos todos los filtros: */
    const ret = enumObject(filtros, (key, value) => ({ ...value, key })).reduce((datos, filtro) => aplicarFiltro(filtro, datos), data);
    return ret;
}

/**Función que se usa para la búsqueda por texto */
export interface TextSearchFunc<T> {
    (pattern: string, item: T, columnas: GetDataColumna<T>[]): boolean;
}


/**Aplica el filtro de la búsqueda por texto */
function filtroBusqueda<T>(items: ItemRowData<T>[], busqueda: string) {
    if (!busqueda) return items;
    return items.filter(x => search(busqueda, x.textSearch));
}
export interface GetDataColumna<T> {
     /**Función que obtiene el contenido de la columna dado el objeto en esta fila. La función toma dos parametros, el primero es la fila
     * y el segundo los datos compartidos por todos los items
    */
    cell: (row: T) => GridCell<any, any>;
    
}
export interface FiltroOrdenColumna<T>  {
    filtro: FiltroFunc<CellData, {}>;
}
