import { GridRow, GridColumn, GridCellExEx, GridRowExEx, ColumnaProps, GridCellEx, GridColumnEx } from "../grid/tipos";
import { toGridRowExEx, filtrarItems } from "../grid/logic"
import { GridCellControls } from "../grid/components"
import { Grid, GridState } from "../grid/grid";
import React = require("react");
import { SelectorValue } from "../selector"
import { Checkbox } from "../fields/checkbox";
import { async } from "../async";
import memoize = require("memoizeref");
import { MakeUncontrolledCookies } from "../HOC";

export interface GridSelectorElement {
    id: SelectorValue;
    row: GridRow;
}

interface SyncMultiSelectGridState {
    allCheckboxValue: boolean;
}

interface SyncMultiSelectGridProps {
    /**Elementos del selector */
    items: GridSelectorElement[];
    /**Columnas del grid */
    columns: GridColumn[];
    value: SelectorValue[];
    onChange?: (value: SelectorValue[]) => void;
    /**Propiedades que se le pasaran a los rows que tiene una función para devolver la fila visual */
    visualProps?: any;

    /**Valores del filtro y del orden.*/
    gridState: GridState;
    /**Se llama al modificar el filtro orden.*/
    onGridStateChange?: (value: GridState) => void;
}

class ControlledSyncMultiSelectGrid extends React.PureComponent<SyncMultiSelectGridProps, SyncMultiSelectGridState> {
    constructor(props) {
        super(props);

        this.state = {
            allCheckboxValue: false
        };
    }

    get itemsFiltrados() {
        //Todos los items filtrados:
        const seleccion = this.props.value;
        //Tenemos que agregar la columna del checkbox para que se comporte bien con los filtros:
        const todosCheck = this.props.items.map(x => ({ id: x.id, row: this.createRow(toGridRowExEx(x.row), x.id, seleccion) }));
        //Se le suma 1 para incluir la columna del checkbox
        const columnCount = this.props.columns.length  + 1;
        const todos = filtrarItems(todosCheck, x => x.row,  this.props.gridState.filtros, this.props.gridState.busqueda, columnCount);
        return todos;
    }


    private handleAllCheckbox = (value: boolean) => {
        if (this.props.onChange) {
            const todos = this.itemsFiltrados;
            const idsFiltrados = new Set(todos.map(x => x.id));
            const viejos = this.props.value;
            if (value) {
                //Agregamos todos los ids a los ids actuales, quitando los repetidos con el set
                const nuevos = [... new Set([...viejos, ...idsFiltrados])];
                this.props.onChange(nuevos);
            } else {
                //Quitamos los ids filtrados:
                const nuevos = viejos.filter(x => !idsFiltrados.has(x));
                this.props.onChange(nuevos);
            }

            this.setState(prev => ({ allCheckboxValue: !prev.allCheckboxValue }));
        }
    }

    private get columns() {
        const allCheck: GridColumnEx =
            {
                header: (props: ColumnaProps) =>
                    <td key="select_all_check">
                        <span>
                            <Checkbox value={this.state.allCheckboxValue} onChange={this.handleAllCheckbox} />
                            <GridCellControls {...props} />
                        </span>
                    </td>,
                titulo: "Selección"
            }
        return [allCheck, ...this.props.columns];
    }

    private handleSelectionChange = (id: SelectorValue, value: boolean) => {
        if (this.props.onChange) {
            const old = this.props.value;
            const exists = old.indexOf(id) != -1;

            const newValue =
                (exists && !value) ? old.filter(x => x != id) :  //quitarlo
                    (!exists && value) ? [...old, id] : //agregarlo
                        old; //queda igual

            this.props.onChange(newValue);
        }
    }

    private createRow(row: GridRowExEx, id: SelectorValue, seleccion: SelectorValue[]): GridRowExEx {
        const creatCheckCell = (selected: boolean, value: SelectorValue) => ({
            data: selected,
            view: <td key="multiselect_checkbox"><Checkbox value={selected} onChange={x => this.handleSelectionChange(value, x)} /></td>
        });

        return {
            props: row.props,
            data: (index: number) => index == 0 ? (seleccion.indexOf(id) != -1) : row.data(index - 1),
            visual: (visualProps: SelectorValue[]) => [creatCheckCell(visualProps.indexOf(id) != -1, id), ...row.visual(this.props.visualProps)],
        };
    }

    private getRows = memoize((items: GridSelectorElement[], seleccion: SelectorValue[]): GridRowExEx[] => {
        const mappedItems = items.map((x, i) => ({ row: toGridRowExEx(x.row), id: x.id }));
        return mappedItems.map(x => this.createRow(x.row, x.id, seleccion));
    });

    private get rows() {
        return this.getRows(this.props.items, this.props.value);
    }

    render() {
        const columns = this.columns;
        const rows = this.rows;
        return (
            <Grid
                columns={columns}
                rows={this.rows}
                visualProps={this.props.value}
                gridState={this.props.gridState}
                onGridStateChange={this.props.onGridStateChange}
            />
        );
    }
}

export interface MultiSelectGridProps {
    /**Elementos del selector */
    items: PromiseLike<GridSelectorElement[]> | GridSelectorElement[];
    /**Columnas del grid */
    columns: PromiseLike<GridColumn[]> | GridColumn[];
    value: PromiseLike<SelectorValue[]> | SelectorValue[];
    onChange?: (value: SelectorValue[]) => void;
    /**Valores del filtro y del orden.*/
    gridState?: GridState;
    /**Se llama al modificar el filtro orden.*/
    onGridStateChange?: (value: GridState) => void;
    /**Nombre del cookie que se usara para almacenar el gridState en caso de que se quiera recordar */
    cookieKey?: string;
}

class ControlledMultiSelectGrid extends React.PureComponent<MultiSelectGridProps, {}> {
    render() {
        return async(Promise.all([this.props.items, this.props.columns, this.props.value]),
            all => <ControlledSyncMultiSelectGrid
                items={all[0]}
                columns={all[1]}
                value={all[2]}
                onChange={this.props.onChange}
                gridState={this.props.gridState || { busqueda: "", filtros: [] }}
                onGridStateChange={this.props.onGridStateChange}
            />)
    }
}

export const MultiSelectGrid = MakeUncontrolledCookies<MultiSelectGridProps>("gridState", "onGridStateChange")(ControlledMultiSelectGrid);