import React = require("react");
import table = require("./table");
import selector = require("./selector");
import { async } from "./async";
import { MakeUncontrolledCookies } from "./HOC"
/**Propiedades del paginador */
export interface PageSelectorProps {
    /**Cantidad de paginas */
    count: number;
    /**Pagina seleccionada, iniciando en 0 */
    selected?: number;
    /**Se ejecuta cuando el usuario de la click a una pagina, el numero de pagina inicia en 0 */
    onValueChanged?: (page: number) => void;
}

/**Un selector de pagina */
export class PageSelector extends React.PureComponent<PageSelectorProps, {}>{
    get Items() {
        var ret: selector.SelectorElement[] = [];
        var paginas = selector.Rango(this.props.count).map(x => ({ value: x, label: x + 1, tooltip: "página " + (x + 1) }));

        return paginas;
    }

    handleOnChange = (value) => {
        if (!this.props.onValueChanged) return;
        this.props.onValueChanged(value);
    }

    render() {
        return (
            <selector.PaginatedButtonSelector
                items={this.Items}
                value={this.props.selected}
                onChange={this.handleOnChange}
            />
        )
    }
}



const commonPageSizes = [10, 15, 30, 50, 100, 200, 500, 1000, 2000, 5000, 10000];

interface PaginatedTableProps {
    onChange?: (value: table.PaginatedTableState) => void;
    value: table.PaginatedTableState;

    /**Cabecera de la tabla. Generalemente esto es un elemento tr */
    header: React.ReactNode;
    /**Elementos de la tabla */
    items?: (JSX.Element | (() => JSX.Element))[];
}

/**Una tabla paginada, el usuario puede cambiar la pagina actual.
 * Establezca los elementos tr de la tabla como los hijos de este componente
*/
class SyncPaginatedTable extends React.PureComponent<PaginatedTableProps>{
    constructor(props) {
        super(props);
    }

    /**Cantidad total de elementos */
    private get ItemsCount() {
        return this.Items.length;
    }

    /**Cantidad de páginas */
    private get PageCount() {
        return Math.ceil(this.ItemsCount / this.PageSize);
    }

    /**Tamaño de la página */
    private get PageSize() {
        return this.props.value.selectedPageSize;
    }

    /**Todos los elementos */
    private get Items() {
        return (this.props.items || []);
    }

    /**Maxima pagina que se puede seleccionar */
    private get MaxPage() {
        return Math.max(this.PageCount - 1, 0);
    }

    /**Pagina que se debe de mostrar como la seleccionada */
    private get EffectiveSelected() {
        return Math.min(this.selectedPage, this.MaxPage);
    }

    private get selectedPage() {
        return this.props.value.selectedPage;
    }

    private get PageItems() {
        const selectedPage = this.EffectiveSelected;
        return (
            this.Items
                .slice(selectedPage * this.PageSize, (selectedPage + 1) * this.PageSize)
                .map(x => (typeof x == "function") ? x() : x)
        );
    }

    /**Tamaños de página disponibles para que el usuario seleccione */
    private get PageSizes() {
        const count = this.ItemsCount;
        var ret = commonPageSizes.filter(x => x < count);
        ret.push(count);
        return ret;
    }

    /**Determina si se deben de mostrar los controles del paginador */
    private get ShowPageControls() {
        const minPageSize = commonPageSizes[0];
        return this.PageCount > 0 && this.Items.length > minPageSize;
    }

    private handleSelectedPageChange = (page: number) => {
        if (this.props.onChange) {
            this.props.onChange({ ... this.props.value, selectedPage: page });
        }
    }

    private handlePageSizeChange = (size: number) => {
        if (this.props.onChange) {
            this.props.onChange({ ... this.props.value, selectedPageSize: size });
        }
    }

    render() {
        //TODO: Usar aqui el PageSelector
        return (
            <div>
                <span>
                    {/**Selector de página*/
                        this.ShowPageControls &&
                        <PageSelector
                            onValueChanged={this.handleSelectedPageChange}
                            selected={this.EffectiveSelected}
                            count={this.PageCount} />
                    }
                    {
                        /**Selector del tamaño de la página*/
                        this.ShowPageControls &&
                        <div className="pull-right">
                            <selector.ButtonSelector
                                items={this.PageSizes.map(x => ({ value: x, label: x, tooltip: x + " elementos por página" }))}
                                value={this.props.value.selectedPageSize}
                                onChange={this.handlePageSizeChange} />
                        </div>
                    }
                </span>
                <table.Table header={this.props.header} items={this.PageItems} />
            </div>
        );
    }
}

/**Tabla paginada con soporte para items y header asíncronos */
class ControlledPaginatedTable extends React.PureComponent<table.AsyncPaginatedTableProps>{
    constructor(props) {
        super(props);
        this.state = { SelectedPage: 0, SelectedPageSize: 15 };

        if (props.children) {
            console.warn("No mande los hijos de la tabla como hijos del elemento, utilice props.items");
        }
    }

    componentWillReceiveProps(props: Readonly<table.AsyncPaginatedTableProps & { children }>) {
        if (props.children) {
            console.warn("No mande los hijos de la tabla como hijos del elemento, utilice props.items");
        }
    }

    render() {
        const prom = Promise.all([this.props.header, this.props.items || (this.props.children as JSX.Element[]) || []]);
        return (
            async(prom, (value) =>
                <SyncPaginatedTable
                    header={value[0]}
                    items={value[1]}
                    value={this.props.value || { selectedPage: 0, selectedPageSize: 15 }}
                    onChange={this.props.onChange}
                />

            )
        )
    }
}

export const PaginatedTable = MakeUncontrolledCookies<table.AsyncPaginatedTableProps>("value", "onChange")(ControlledPaginatedTable);