import React = require('react');
import cargando = require('./cargando');
import promiseEx = require("./promiseEx");
/**Interfaz base para los estados de un catalogo */
export interface QueryState<T> {
    /**Resultado del query*/
    Model?: T | null;
    /**True si se estan cargando los elementos */
    Cargando?: boolean;
    /**Si hubo un error al cargar se mostrará aqui */
    Error?: string | number | null;
}


export interface ModelMapFunction<T> {
    (model: T, onModelChanged: (newValue: T) => void, requestQuery: () => void): JSX.Element | null;
}

export interface CargandoMapFunction<T> {
    (cargando: boolean, error: string | number | null | undefined, model: T | null | undefined, onModelChanged: (newValue: T) => void, requestQuery: () => void): JSX.Element | null;
}

function verifyFunction(func: any) {
    if (func == null || typeof func != "function")
        throw "El argumento 'map' debe de ser una función";
}

/**Convierte una función que mapea un modelo a un componente de react a otra que mapea el modelo a un componente 
 * de cargando que muestra el error y el spinner de cargando. La función que devuelve es compatibla con la función que espera
 * el componente Query
 */
export function CargandoModelMap<T>(map: ModelMapFunction<T>): CargandoMapFunction<T> {
    verifyFunction(map);

    return (_cargando: boolean, error: string | number | null, model: T | null | undefined, onModelChanged, requestQuery) =>
        <cargando.Cargando Error={error} Cargando={_cargando} >
            {!error && !_cargando && map(model!, onModelChanged, requestQuery)}
        </cargando.Cargando>;
}

/**Convierte una función que mapea un modelo a un componente de react a otra que mapea el modelo a un
 * componente de cargando con su respectivos elementos constantes de cargando y una función para crear el componente de error
  */
export function ConstantModelMap<T>(map: ModelMapFunction<T>, cargandoComponent: JSX.Element | null, errorFunc: (error: string | number) => JSX.Element | null) {
    verifyFunction(map);

    return (_cargando: boolean, error: string | number | null, model: T | null , onModelChanged, requestQuery) =>
        error ? errorFunc(error) :
            _cargando ? cargandoComponent : map(model!, onModelChanged, requestQuery);
}

/**Propiedades de un elemento que realiza un query */
export interface QueryProps<T, TQueryParams> {
    /**Función que realiza el query */
    query: (params: TQueryParams) => PromiseLike<T> | T;
    /**Función que convierte el resultado del query en un elemento de react.
     * 
     * @param model Resultado del query.
     * @param onModelChanged Delegado que debe de ser llamado cuando se quiera informar que el modelo ha cambiado
     * @param requestQuery Delegado que debe de ser llamado si se debe de llamar de nuevo al query
     */
    map: CargandoMapFunction<T>;

    /**Parámetros del query */
    queryParams: TQueryParams;
    /**True para indicar que no se debe de llamar al refrescado al iniciar */
    preventInitialRequest?: boolean;
}



/**Componente que muestra un componente en función al resultado de un query asíncrono que esta 
 * en función de ciertos parametros
 * debera de heredar del componente para establecer los tipos genericos.
 * 
 * Cuando cambien los parametros del query automaticamente vuelve a llamar al query
 */
export class Query<T, TQueryParams>
    extends React.Component<QueryProps<T, TQueryParams>, QueryState<T>>{
    constructor(props: QueryProps<T, TQueryParams>) {
        super(props);
        console.warn("El componente kea.Query esta obsoleto, utilice kea.Async");
        //Initial state:
        this.state = {
            Cargando: !props.preventInitialRequest
        };
    }

    componentDidMount() {
        if (!this.props.preventInitialRequest)
            this.refreshFull();
    }
    componentWillReceiveProps(nextProps: QueryProps<T, TQueryParams>) {
        //Si cambiaron los parametros del query, refrescamos
        if (JSON.stringify(this.props.queryParams) != JSON.stringify(nextProps.queryParams)) {
            this.refreshFull();
        }
    }
    /**Refresca el query */
    refreshFull = async () => {
        this.setState({ Cargando: true, Error: null });
        try {
            //Delay para forzar la actualización del estado anterior
            await promiseEx.Delay(1000);
            var result = await this.props.query(this.props.queryParams);

            this.setState({
                Model: result,
                Cargando: false,
                Error: null
            });
        }
        catch (err) {
            console.error("Query error: " + err);
            this.setState({
                Model: null,
                Cargando: false,
                Error: err
            });
        }
    }

    private handleOnModelChanged = (newValue: T) => {
        this.setState({ Model: newValue });
    };
    render() {
        return this.props.map(
            this.state.Cargando || false,
            this.state.Error,
            this.state.Model,
            this.handleOnModelChanged,
            this.refreshFull
        );
    }
}

/**Query tipado con any */
export class ConcreteQuery extends Query<any, any>  { } 