import React = require("react");
import interfaces = require("./interfaces");
import * as ReactSelect  from "react-select";
import { async } from "./async";
import { FormField } from "./fields/formField";
import { Text } from "./fields/text";
import { search } from "./search";
import { Observable } from "rxjs";
import { componentToRx } from "react-rxk";

/**Genera un arreglo de numeros consecutivos iniciando en el 0 */
export function Rango(count: number): number[] {
    var r: number[] = [];
    for (var i = 0; i < count; i++) {
        r.push(i);
    }
    return r;
}

export type SelectorValueNotNull = string | number | boolean | Date;
export type SelectorValue = SelectorValueNotNull | undefined | null;
export type SelectorLabel = string | number | undefined | null;
/**Un elemento de un selector */
export interface SelectorElement {
    /**Valor del elemento. El indefinido sólo es válido si disabled == true */
    value?: SelectorValue;
    label: SelectorLabel;
    tooltip?: SelectorLabel;
    disabled?: boolean;
    style?: "danger" | "warning" | "success" | "info"
}

/**Propiedades de un selector de botones */
export interface SelectorProps extends interfaces.ControlledValueProps<SelectorValue, SelectorValue>, interfaces.FormFieldDefaultProps {
    /**Elementos del selector*/
    items?: SelectorElement[];

    error?: boolean;
}

export interface AsyncSelectorProps extends interfaces.AsyncControlledValueProps<SelectorValue, SelectorValue>, interfaces.FormFieldDefaultProps {
    /**Elementos del selector*/
    items?: SelectorElement[] | PromiseLike<SelectorElement[]>;
    error?: boolean;
    /**True para colorear de rojo el control si no hay ningun valor seleccionado */
    required?: boolean;
}
/**Componente que muestra un selector con botones */
export class ButtonSelector extends React.PureComponent<SelectorProps, {}> {
    private onClick(value: SelectorValue) {
        if (this.props.onChange)
            this.props.onChange(value);
    }

    render() {
        return (
            <ul className="pagination pagination-sm">
                {
                    (this.props.items || [])
                        .map((x, i) =>
                            <li key={x.value == null ? ("__null" + i) : (x.value + "")}
                                className={
                                    x.disabled ? "disabled" :
                                        this.props.value == x.value ? "active" : ""}>

                                <a
                                    /**Arreglo del z index, el selector activo sale por encima de los drop downs de los combos */
                                    style={{ zIndex: 0 }}
                                    title={x.tooltip ? (x.tooltip + "") : ""}
                                    onClick={x.disabled ? undefined : (() => this.onClick(x.value))} >
                                    {x.label}
                                </a>

                            </li>)
                }
            </ul>
        );
    }
}



const reactSelectFix = {
    menuContainerStyle: { 'zIndex': 999 },
    filterOption: (option, filter) => search(filter, option.label) as any
}


/**Selector simple */
class SelectBase extends React.PureComponent<AsyncSelectorProps, {}> {
    handleOnChange = (option: ReactSelect.Option | null) => {
        if (this.props.onChange) {
            this.props.onChange(option && (this.unmapValue(option.value)))
        }
    }

    mapValue(value: SelectorValue): string | number | undefined {
        if (value == 0)
            return "__NUMBERZERO__";
        else if (typeof (value) == "string" || typeof (value) == "number") {
            return value;
        } else if (typeof (value) == 'boolean') {
            return "__BOOLEAN__" + value;
        } else if (value === null) {
            return "__NULL__";
        } else if (value instanceof Date) {
            return "__DATE__" + value.toISOString();
        } else {
            return value;
        }
    }

    unmapValue(value: string | number | undefined | boolean): SelectorValue {
        if (value === "__BOOLEAN__" + true)
            return true;
        else if (value === "__BOOLEAN__" + false)
            return false;
        else if (value === "__NULL__")
            return null;
        else if (value === "__NUMBERZERO__")
            return 0;
        else if (typeof value == "string" && value.startsWith("__DATE__")) {
            const part = value.substr("__DATE__".length);
            return new Date(part);
        }
        else
            return value;
    }
    private async selectedText() {
        const items = await (this.props.items || []);
        const value = await this.props.value;
        const selected = items.filter(x => x.value == value)[0];
        return (selected && (selected.label + "")) || "(ninguno)";
    }

    private async mappedValue() {
        return this.mapValue(await this.props.value);
    }

    private async items() {
        if (this.props.items) {
            const mapItem: (item: SelectorElement) => ReactSelect.Option =
                x => ({ value: this.mapValue(x.value), label: "" + x.label, className: x.style ? "alert-" + x.style : undefined });

            return (await this.props.items).map(mapItem);
        } else return [];
    }

    render() {
        //En lugar de usar el kea.async para resolver la promesa de los items se la delegamos al ReactSelect para que se cargue mas 
        //rapido la interfaz de usuario cuando aún no estan disponibles las listas
        const selectedText = this.selectedText();
        type SelectValue = string | number | undefined;
        const getSelect = (value: SelectValue, options: ReactSelect.Option[], cargando: boolean) =>
            <ReactSelect.default
                placeholder="Seleccionar..."
                noResultsText="No hay resultados."
                options={options}
                isLoading={cargando}
                value={value}
                onChange={this.handleOnChange}
                {...reactSelectFix}
            />;

        const realSelect = (value: SelectValue, options: ReactSelect.Option[]) => getSelect(value, options, false);
        const cargandoSelect = getSelect(undefined, [], true);

        const error = this.props.error || ((!!this.props.required) && (this.props.value == null));
        return (
            <FormField label={this.props.label} error={error} >
                {
                    this.props.isReadOnly ?
                        <Text value={this.selectedText()} isReadOnly={true} /> :
                        async(this.mappedValue(), value =>
                            async(this.items(), items => realSelect(value, items), { cargando: cargandoSelect })
                        )
                }
            </FormField>
        );
    }
}
/**Selector simple */
export const Select = componentToRx(SelectBase);
/**Props de un selector múltiple */
export interface MultiSelectProps extends interfaces.ControlledValueProps<SelectorValue[], SelectorValue[]> {
    items?: SelectorElement[] | PromiseLike<SelectorElement[]>;
    label?: JSX.Element | string;
    readonly?: boolean;
}

/**Selector múltiple. El valor seleccionado es un arreglo de los valores de los elementos */
export class MultiSelect extends React.PureComponent<MultiSelectProps, {}> {
    handleOnChange = (options: ReactSelect.Option[] | ReactSelect.Option | null) => {
        if (this.props.onChange) {
            this.props.onChange(options && (options as ReactSelect.Option[]).map(x => x.value!) || [])
        }
    }

    private async  items() {
        if (this.props.items) {
            return (await this.props.items).map(x => {
                return ({ value: x.value, label: "" + x.label } as ReactSelect.Option)
            })
        } else return [];
    }


    render() {
        /*
        Hay un problema con el ReactSelect multiple y asíncronom asi que en este caso usaremos el kea.async en lugar de delegar la promesa al react select
        https://github.com/JedWatson/react-select/issues/1542
        */
        const value = this.props.value == null ? [] : (this.props.value as string[]);
        const getSelect = (options: ReactSelect.Option[], cargando: boolean) =>
            <ReactSelect.default
                inputProps={{ readOnly: this.props.readonly || undefined }}
                placeholder="Seleccionar..."
                noResultsText="No hay resultados."
                options={options}
                isLoading={cargando}
                value={value}
                multi={true}
                onChange={this.props.readonly ? undefined : this.handleOnChange}

                {...reactSelectFix}
            />;

        const realSelect = (options: ReactSelect.Option[]) => getSelect(options, false);
        const cargandoSelect = getSelect(value.map(x => ({ value: x, label: x } as ReactSelect.Option)), true);

        return (
            <FormField label={this.props.label}>
                {
                    async(this.items(), items => realSelect(items), { cargando: cargandoSelect })
                }
            </FormField>
        );
    }
}

/**Componente que muestra un selector de botones paginado, útil cuando existen demasiadas opciones  */
export class PaginatedButtonSelector extends React.PureComponent<SelectorProps, {}> {
    constructor(props) {
        super(props);
    }

    PageSize = 10;

    /**Todos los items */
    get Items() {
        return this.props.items || [];
    }

    handleOnChange = (value: SelectorValue) => {
        if (!this.props.onChange) return;
        const index = this.CurrentIndex || 0;
        const items = this.Items;
        if (value == "_next") {
            this.props.onChange(this.GetValueAt(Math.min(items.length - 1, index + 1)));
        }
        else if (value == "_previous") {
            this.props.onChange(this.GetValueAt(Math.max(0, index - 1)));
        } else {
            this.props.onChange(value);
        }
    }

    GetValueAt(index: number) {
        return this.Items[index].value;
    }

    get CurrentIndex() {
        return this.GetIndex(this.props.value);
    }

    GetIndex(value: SelectorValue) {
        const items = this.Items;
        for (var i = 0; i < items.length; i++) {
            if (items[i].value == value)
                return i;
        }
        return null;
    }

    /**Elementos de la página actual */
    get PageItems() {
        const pageCount = this.PageCount;
        const items = this.Items;
        //Si no hay paginas, devolvemos items, que seria un arreglo vacio
        if (pageCount <= 1)
            return this.Items;

        const page = this.CurrentPage;
        const pageItems = this.Items.slice(this.CurrentPage * this.PageSize, (this.CurrentPage + 1) *
            this.PageSize);
        const pageSize = this.PageSize;

        //Agregamos los puntos suspensivos al principio o al final de la lista:
        const dots: SelectorElement = { value: undefined, label: "...", disabled: true };
        if (page > 0) {
            const lastPageItem = (page - 1) * pageSize;
            pageItems.splice(0, 0, dots);
            if (lastPageItem != 0)
                pageItems.splice(0, 0, items[lastPageItem]);
            pageItems.splice(0, 0, items[0]);
        }

        if (page < (pageCount + 1)) {
            const lastItem = items.length - 1;
            const nextPageItem = Math.min(lastItem, (page + 2) * pageSize - 1);
            pageItems.push(dots);
            if (nextPageItem != lastItem)
                pageItems.push(items[nextPageItem]);
            pageItems.push(items[items.length - 1]);
        }

        pageItems.splice(0, 0, { value: "_previous", label: '«' });
        pageItems.push({ value: "_next", label: '»' });

        return pageItems;
    }

    /**Indice seleccionado */
    get SelectedIndex(): number | null {
        var index = this.CurrentIndex;
        if (index === undefined) return null;
        else return index;
    }

    /**Pagina actual */
    get CurrentPage() {
        return Math.floor((this.SelectedIndex || 0) / this.PageSize);
    }

    /**Cantidad de páginas */
    get PageCount() {
        return Math.ceil(this.Items.length / this.PageSize);
    }

    render() {
        return (
            <ButtonSelector
                items={this.PageItems}
                value={this.props.value}
                onChange={this.handleOnChange}
            />
        )
    }
}

export interface NeastedSelectorElement extends SelectorElement {
    parentValue: SelectorValue;
}


/*Propiedades de un selector anidado */
export interface NeastedSelectProps {
    /**Padre seleccionado en función del hijo seleccionado */
    parentValue: SelectorValue;

    /**Función que dibuja el elemento basado en el valor padre seleccionado */
    map: (parentValue: SelectorValue, onParentChange: (newValue: SelectorValue) => void) => JSX.Element | null;

    /**Cuando cambie el padre seleccionado, esta función se llamara con un valor de null para resetear el hijo seleccionado*/
    onChildChange?: (value: null) => void;
}

export interface NeastedSelectState {
    parentValue?: SelectorValue;
}

/**Un selector anidado */
export class NeastedSelector extends React.PureComponent<NeastedSelectProps, NeastedSelectState> {
    constructor(props) {
        super(props);
        this.state = {};
    }

    handleParentChange = (value: SelectorValue) => {
        if (this.props.onChildChange == null) return;
        //Si cambia el padre, deseleccionamos el hijo
        this.props.onChildChange(null);
        this.setState({ parentValue: value });
    }

    get PadreActual() {
        if (this.props.parentValue == null)
            return this.state.parentValue;
        else
            return this.props.parentValue;
    }

    render() {
        return this.props.map(this.PadreActual, this.handleParentChange);
    }
}