import React = require('react');
import { BootstrapButtonStyle, Button } from "./botones";

export interface UploadProps {
    onChange?: (value: File[]) => Promise<void> | void;

    disabled?: boolean;
    tooltip?: string;
    /**True para seleccionar multiples archivos */
    multiple?: boolean;
    /**Estilo del botón de bootstrap. Por default es "default" */
    style?: BootstrapButtonStyle;
    /**Clase del icono o indefinido para no mostrar un icono */
    iconClass?: string;
}

function hideInput(el: HTMLElement) {
    el.style.position = 'absolute';
    el.style.top = "0";
    el.style.left = "0";
    el.style.width = "0";
    el.style.height = "0";
    el.style.display = "none"
}

/**Obtiene un arreglo con los archivos seleccionados en un input. Devuelve null si no se seleccionó ningun archivo */
function getFiles(input: HTMLInputElement): File[] | null {
    if (!input.files || input.files.length == 0) return null;
    const multi = input.multiple;
    const inputFiles = input.files;

    const fileArray: File[] = [];
    for (let i = 0; i < inputFiles.length; i++) {
        fileArray.push(inputFiles[i]);
    }

    return fileArray;
}

/**Muestra el selector de archivos y devuelve el archivo seleccionado o null si no hay archivo */
export function upload(): Promise<File | null>
/**Muestra el selector de archivos y devuelve el archivo seleccionado o null si no hay archivo */
export function upload(multi: false): Promise<File | null>
/**Muestra el selector de archivos y devuelve los archivos seleccionados o null si no hay archivos */
export function upload(multi: true): Promise<File[] | null>
export function upload(multi?: boolean): Promise<File | File[] | null> {
    //Creamos un elemento input escondido:
    const input = document.createElement("input");
    input.type = "file";
    hideInput(input);
    input.multiple = multi || false;
    //Lo agregamos al documento:
    document.body.appendChild(input);
    return new Promise(resolve => {
        input.onchange = (ev) => {
            const files = getFiles(input);
            const ret = files && (multi ? files : files[0]);
            //Quitamos el input del documento:
            document.body.removeChild(input);
            resolve(ret); 
        };

        /**Bandera que indica si se debe de revisar si el input se cancelo, esto porque sólo se debe de hacer una vez */
        let checkCancel = true;
        function onBodyFocus() {
            if(!checkCancel) 
                return;
            checkCancel= false;

            document.body.removeEventListener("focus", onfocus);

            //Realizamos el chequeo con un retraso, si lo revisamos inmediatamente el valor de input.value aun no
            //estará actualizado
            setTimeout(() => {
                if(!input.value) {
                    //El usuario no seleccionó ningún archivo
                    resolve(null);
                }
            }, 100);
        }
        document.body.addEventListener("focus", onBodyFocus, true);

        input.click();
    });

}


/**Esta clase sirve para mostrar un botón que nos permite abrir un filechooser y regresa un arreglo de tipo File
 * @deprecated Utilice en lugar de este componente un Button y la función upload()
 */
export class Upload extends React.Component<UploadProps, {}> {
    constructor(props) {
        super(props);
        console.warn("@deprecated Utilice en lugar del componente <Upload /> un <Button /> y la función upload()");
    }
    
    input: HTMLInputElement;

    handleInput = (input: HTMLInputElement | null) => {
        if (input) {
            this.input = input;
        }
    }

    handleOnChangeAsync = async (files: FileList | null) => {
        if (!this.props.onChange) return;
        if (!files || files.length == 0) return;

        const ret: File[] = [];
        for (let i = 0; i < files.length; i++) {
            ret.push(files.item(i));
        }

        await this.props.onChange(ret);
    }

    handleOnChange = (a: React.ChangeEvent<HTMLInputElement>) => {
        this.handleOnChangeAsync(this.input!.files);

        this.input!.value = null as any;
        return false;
    }

    handleOnClick = () => {
        this.input!.click();
    }

    render() {
        const esconder: React.CSSProperties = {
            position: 'absolute',
            top: 0,
            left: 0,
            width: 0,
            height: 0,
            display: "none"
        };
        return (
            <div style={{ position: "relative", backgroundColor: "white" }}>
                <input type="file" onChange={this.handleOnChange} style={esconder} multiple={this.props.multiple} ref={this.handleInput} />
                <Button
                    onClick={this.handleOnClick}
                    disabled={this.props.disabled}
                    iconClass={this.props.iconClass || "fa fa-upload"}
                    style={this.props.style || "info"}
                    tooltip={this.props.tooltip}
                >
                    {this.props.children || "Buscar archivo"}
                </Button>
            </div>
        );
    }
}

