import { enumObject, any, mapObject, pipe, union, arrayToMap, promiseAllObj } from "keautils";
export type ValidationRuleResult = boolean | string | null | undefined;
/**Una función de validación. Devuelve false o una cadena no vacia si hay un error de validación, devuelve true si no hay error de validacón*/
export interface Rule<T> {
    (value: T, key: string): PromiseLike<ValidationRuleResult> | ValidationRuleResult;
}

export type Rules<T> = {
    [K in keyof T]?: Rule<T>
}

/**Contiene los mensajes de errores de las propiedades que tienen error */
export type Result<T> = {
    [K in keyof T]?: Promise<string | undefined>;
}

function resultIsError(value: ValidationRuleResult): boolean {
    if (!value) return true;
    if (value == true) return false;
    return true;
}

/**Determina si un resultado de validación tiene errores */
export async function hasError<T>(validation: Result<T>) {
    const val = await promiseAllObj(validation);
    const ret = any(enumObject(val), x => x.value != null);
    return ret;
}

/**Evalua un conjunto de reglas de validación dado un objeto */
export function getErrors<T>(rules: Rules<T>, value: T): Result<T> {
    const result = pipe(
        rules,
        curr => mapObject(curr, (prop, key) => prop ? prop(value, key) : true),
        curr => mapObject(curr, async (prop) => {
            const pVal = await prop;
            return !(pVal) ? "error" : //Si prop evalua a falso, es error
                (pVal) === true ? undefined : //Si prop === true, no es error
                    (pVal);
        }
        )
    );

    return result;
}



/**Devuelve la cadena "Valor requerido" si value es null o undefined, si no devuelve true*/
export function required(value: any, key: string) {
    const val = (value && value[key])
    return (val == null || val == "") ? "Valor requerido" : true;
}

/**
 * Devuelve un nuevo conjunto de reglas el cual sólo valida a errores cuando se cumple la condición, es decir, si la condición no se cumple, el nuevo conjunto de reglas no devuelve errores
 * @param rules Conjunto de reglas base
 * @param condition Condición tal que sólo permitirá devulver errores si es true
 */
export function when<T>(condition: ((value: T) => boolean) | boolean, rules: Rules<T>) {
    if (condition === true) {
        return rules;
    }
    else if (condition === false) {
        return {};
    }
    else {
        return mapObject(rules, (rule) => rule && ((value, key) => condition(value) ? rule(value, key) : true))
    }
}

/**
 * Devuelve un nuevo conjunto de reglas que es el AND entre dos conjuntos de reglas
 * @param a 
 * @param b 
 */
function and2<T>(a: Rules<T>, b: Rules<T>): Rules<T> {
    function resultAnd(a: ValidationRuleResult, b: ValidationRuleResult): ValidationRuleResult {
        if (a === true && b === true) return true;
        if (typeof a == "string" && typeof b == "string") return a + ", " + b;
        if (typeof a == "string") return a;
        if (typeof b == "string") return b;

        else return a || b;
    }

    function ruleAnd(a: Rule<T>, b: Rule<T>) {
        return async (value: T, key: string) => {
            const aProm = Promise.resolve(a(value, key));
            const bProm = Promise.resolve(b(value, key));
            const race = await Promise.race([aProm, bProm]);
            if (resultIsError(race))
                return race;
            else
                return resultAnd(await aProm, await bProm);
        }
    }

    const keys = union(Object.keys(a), Object.keys(b));
    const values = arrayToMap(keys.map((key: keyof T) => {
        const aRule = a[key];
        const bRule = b[key];

        const value =
            aRule && bRule ? ruleAnd(aRule, bRule) :
                aRule || bRule;
        return { key, value };
    }));

    return values as Rules<T>;
}

/**
 * Devuelve un nuevo conjunto de reglas que es el AND entre dos conjuntos de reglas
 */
export function and<T>(...rules: Rules<T>[]): Rules<T> {
    return rules.reduce(and2);
}